from .type import XpdtType

__all__ = (
    'UuidType',
)


class UuidType(XpdtType):
    __slots__ = ()

    @property
    def ctype(self) -> str:
        return 'xu128_t'

    @property
    def pytype(self) -> str:
        return 'bytes'

    @property
    def struct_fmt(self) -> str:
        return '16s'

    @property
    def size(self) -> int:
        return 16

    @property
    def needs_vbuf(self) -> bool:
        return False

    @property
    def needs_decode(self) -> bool:
        return False



from typing import NamedTuple

from .type import XpdtType
from .typedef import TypeDef


__all__ = (
    'MemberDef',
)


class MemberDef(NamedTuple):
    name: str
    typedef: TypeDef

    @property
    def needs_vbuf(self) -> bool:
        return self.typedef.type.needs_vbuf

    @property
    def cdecl(self) -> str:
        return f'{self.typedef.type.ctype} {self.name}'

    @property
    def const_cdecl(self) -> str:
        return f'const {self.cdecl}'

    @property
    def type(self) -> XpdtType:
        return self.typedef.type

    @property
    def is_scalar(self) -> bool:
        return self.typedef.type.is_scalar

    @property
    def is_reserved(self) -> bool:
        return self.name.startswith('_')



from typing import Any, ClassVar, Optional, TypeVar
from pathlib import Path

import logging
import toml

_log = logging.getLogger(__package__)


__all__ = (
    'Registry',
)


_R = TypeVar('_R', bound='Registry')


class Registry:
    MAX_MOD_ID: ClassVar[int] = 0xffff
    MAX_STRUCT_ID: ClassVar[int] = 0xffff

    __slots__ = (
        '_reg',
        '_dirty',
        '_corrupt',
        '_meta',
        '_modules',
        '_module',
    )

    _reg: dict[str, Any]
    _corrupt: bool
    _dirty: bool

    def __init__(self, reg: dict[str, Any]) -> None:
        self._reg = reg
        self._corrupt = False
        self._dirty = False

        self._meta = self._top_level_table('meta')
        self._modules = self._top_level_table('modules')
        self._module = self._top_level_table('module')

    @property
    def corrupt(self) -> bool:
        return self._corrupt

    @property
    def dirty(self) -> bool:
        return self._dirty

    def _create_toplevel(self, table: str) -> dict[str, Any]:
        ret: dict[str, Any] = dict()
        self._reg[table] = ret
        return ret

    def _top_level_table(self, table: str) -> dict[str, Any]:
        """
        Fetch a top-level table from the registry. If it doesn't exist or is
        not a table, then return an empty dict.
        """
        tab = self._reg.get(table)

        if tab is None:
            return self._create_toplevel(table)

        if not isinstance(tab, dict):
            _log.warn('registry.toml: "%s" is not a table', table)
            self._corrupt = True
            return self._create_toplevel(table)

        return tab

    @property
    def name(self) -> Optional[str]:
        meta = self._meta
        name = meta.get('name')
        if not isinstance(name, str):
            return None
        return name

    def _alloc_module_id(self) -> int:
        all_ids = [x for x in self._modules.values()
                   if isinstance(x, int)]
        if not all_ids:
            return 0

        ret = max(all_ids) + 1

        if ret > self.MAX_MOD_ID:
            _log.warn('registry.toml: module ID "%d" is beyond max allowable',
                      ret)
            self._corrupt = True

        return ret

    def _new_module_id(self, module_name: str) -> int:
        mod_id = self._alloc_module_id()
        self._modules[module_name] = mod_id
        self._dirty = True
        return mod_id

    def module_id(self, module_name: str) -> int:
        modules = self._modules
        mod_id = modules.get(module_name)

        if mod_id is None:
            return self._new_module_id(module_name)

        if not isinstance(mod_id, int):
            _log.warn('registry.toml: "modules.%s" is not an int', module_name)
            self._corrupt = True
            return self._new_module_id(module_name)

        return mod_id

    def _new_module_tab(self, module_name: str) -> dict[str, Any]:
        ret: dict[str, Any] = dict()
        self._module[module_name] = ret
        self._dirty = True
        return ret

    def _get_module_tab(self, module_name: str) -> dict[str, Any]:
        module = self._module
        mod_tab = module.get(module_name)
        if mod_tab is None:
            return self._new_module_tab(module_name)

        if not isinstance(mod_tab, dict):
            _log.warn('registry.toml: "module.%s" is not a table', module_name)
            self._corrupt = True
            return self._new_module_tab(module_name)

        return mod_tab

    def _alloc_struct_id(self, mod_tab: dict[str, Any]) -> int:
        all_ids = [x for x in mod_tab.values()
                   if isinstance(x, int)]
        if not all_ids:
            return 0
        ret = max(all_ids) + 1

        if ret > self.MAX_STRUCT_ID:
            _log.warn('registry.toml: struct ID "%d" is beyond max allowable',
                      ret)
            self._corrupt = True
        return ret

    def _new_struct_id(self, mod_tab: dict[str, Any], struct_name: str) -> int:
        struct_id = self._alloc_struct_id(mod_tab)
        mod_tab[struct_name] = struct_id
        self._dirty = True
        return struct_id

    def struct_id(self, module_name: str, struct_name: str) -> int:
        mod_tab = self._get_module_tab(module_name)
        ret = mod_tab.get(struct_name)

        if ret is None:
            return self._new_struct_id(mod_tab, struct_name)

        if not isinstance(ret, int):
            _log.warn('registry.toml: "module.%s.%s" is not an int',
                      module_name, struct_name)
            self._corrupt = True
            return self._new_struct_id(mod_tab, struct_name)

        return ret

    @classmethod
    def empty(cls: type[_R]) -> _R:
        return cls({})

    @classmethod
    def from_file(cls: type[_R], p: Path) -> _R:
        return cls(toml.load(p))

    def save(self, p: Path) -> None:
        if self._dirty:
            with p.open('w') as f:
                toml.dump(self._reg, f)
            self._dirty = False



from .integraltype import IntegralType

__all__ = (
    'IntStackType',
)


class IntStackType(IntegralType):
    __slots__ = (
        '_width',
    )

    _struct_map = {
        8: 'b',
        16: 'h',
        32: 'i',
        64: 'q',
    }

    def __init__(self, width: int):
        super().__init__(64, False)
        self._width = width

    @property
    def _mask(self) -> int:
        return (1 << self._width) - 1

    @property
    def pytype(self) -> str:
        return '_Tup[int]'

    def read_func(self, s: str) -> str:
        return f'_decode_intstack({s}, 0x{self._mask:x}, {self._width})'

    def write_func(self, s: str) -> str:
        return f'_encode_intstack({s}, 0x{self._mask:x}, {self._width})'

    @property
    def needs_decode(self) -> bool:
        return True



from typing import Tuple, Type, TypeVar, NamedTuple, Generator
from enum import Enum

import re


class Case(Enum):
    DASHY = 'dashy'
    SNAKE = 'snake'
    SCREAMING_SNAKE = 'screaming-snake'
    PASCAL = 'pascal'
    CAMEL = 'camel'


T = TypeVar('T', bound='Symbol')


_camel_re = re.compile('(?<=[a-z])(?=[A-Z])')


class Symbol(NamedTuple):
    components: Tuple[str, ...]

    @property
    def dashy(self) -> str:
        return '-'.join(self.components)

    @property
    def snake(self) -> str:
        return '_'.join(self.components)

    @property
    def screaming_snake(self) -> str:
        return '_'.join((c.upper() for c in self.components))

    @property
    def pascal(self) -> str:
        return ''.join((c.title() for c in self.components))

    @property
    def camel(self) -> str:
        p = self.components

        def gen() -> Generator[str, None, None]:
            yield p[0]
            yield from (c.title() for c in p[1:])

        return ''.join(gen())

    @classmethod
    def from_dashy(cls: Type[T], s: str) -> T:
        # TODO: Validate
        return cls(tuple((c.lower() for c in s.split('-'))))

    @classmethod
    def from_snake(cls: Type[T], s: str) -> T:
        # TODO: Validate
        return cls(tuple((c.lower() for c in s.split('_'))))

    @classmethod
    def from_screaming_snake(cls: Type[T], s: str) -> T:
        # TODO: Validate
        return cls(tuple((c.lower() for c in s.split('_'))))

    @classmethod
    def from_pascal(cls: Type[T], s: str) -> T:
        # TODO: Validate
        return cls(tuple((c.lower() for c in _camel_re.split(s))))

    @classmethod
    def from_camel(cls: Type[T], s: str) -> T:
        # TODO: Validate
        return cls(tuple((c.lower() for c in _camel_re.split(s))))

    @classmethod
    def detect(cls: Type[T], s: str) -> T:
        if not s:
            raise ValueError('Cannot have an empty symbol')

        used = frozenset(s)

        begins_lower = s[0].islower()
        has_upper = any(x.isupper() for x in used)
        has_lower = any(x.islower() for x in used)
        has_unders = '_' in used
        has_dashes = '-' in used

        if has_dashes:
            return cls.from_dashy(s)
        elif has_unders:
            if has_upper:
                return cls.from_screaming_snake(s)
            else:
                return cls.from_snake(s)
        elif has_upper and has_lower:
            if begins_lower:
                return cls.from_camel(s)
            else:
                return cls.from_pascal(s)
        elif has_lower:
            return cls.from_snake(s)
        elif has_upper:
            return cls.from_screaming_snake(s)
        else:
            raise ValueError('cannot recognize')


def main() -> None:
    examples = (
        'foo-bar',
        'foo_bar',
        'FOO_BAR',
        'FooBar',
        'fooBar',
        'foo',
        'FOO',
        'Foo',
    )

    for x in examples:
        s = Symbol.detect(x)
        print(s.dashy, s.snake, s.screaming_snake, s.pascal, s.camel)


if __name__ == '__main__':
    main()



import unittest

from io import StringIO

from xpdt import NameSpace, parse, ParseError


_all_fixed = '''
struct fixed {
    u128 u128;
    u64 u64;
    i64 i64;
    u32 u32;
    i32 i32;
    u16 u16;
    i16 i16;
    u8 u8;
    i8 i8;
};
'''

_all_types = '''
struct fixed {
    bytes bytes;
    u128 u128;
    u64 u64;
    i64 i64;
    u32 u32;
    i32 i32;
    u16 u16;
    i16 i16;
    u8 u8;
    i8 i8;
};
'''

class Test_Parser(unittest.TestCase):
    def test_parse_error(self):
        with self.assertRaises(ParseError):
            decls = list(parse('flibble'))

    def test_parse_error_message(self):
        try:
            decls = list(parse('struct x {\nflibble', file='flibble'))
        except ParseError as e:
            buf = str(e)
            self.assertTrue('flibble:2' in buf)

    def test_parse_error_fields(self):
        try:
            decls = list(parse('struct x {\nflibble', file='flibble'))
        except ParseError as e:
            self.assertEqual(e.tok.file, 'flibble')
            self.assertEqual(e.tok.line, 2)

    def test_parse_all_fixed(self):
        decls = parse(_all_fixed)
        ns = NameSpace.from_decls(decls, name=None)
        self.assertEqual(ns.name, None)
        self.assertEqual(len(ns), 1)

    def test_parse_all_types(self):
        decls = parse(_all_types)
        ns = NameSpace.from_decls(decls, name=None)
        self.assertEqual(ns.name, None)
        self.assertEqual(len(ns), 1)

    def test_gen_c_fixed(self):
        decls = parse(_all_fixed)
        ns = NameSpace.from_decls(decls, name=None)
        buf = StringIO()
        ns.gen_c(buf)
        c_prog = buf.getvalue()

    def test_gen_c_vbuf(self):
        decls = parse(_all_types)
        ns = NameSpace.from_decls(decls, name=None)
        buf = StringIO()
        ns.gen_c(buf)
        c_prog = buf.getvalue()

    def test_gen_python_fixed(self):
        decls = parse(_all_fixed)
        ns = NameSpace.from_decls(decls, name=None)
        buf = StringIO()
        ns.gen_python(buf)
        c_prog = buf.getvalue()

    def test_gen_python_vbuf(self):
        decls = parse(_all_types)
        ns = NameSpace.from_decls(decls, name=None)
        buf = StringIO()
        ns.gen_python(buf)
        c_prog = buf.getvalue()



from __future__ import annotations
from typing import NamedTuple, Dict, Generator, List, Optional, Union
from pathlib import Path

from .decl import StructDecl, MemberDecl
from .shiftreduce import state, State, ParseError
from .lex import lex, Lexeme

__all__ = (
    'parse',
    'parse_file',
)


class MemberTypeItem(NamedTuple):
    member_type: str


class MemberNameItem(NamedTuple):
    member_name: str


class NamespaceItem(NamedTuple):
    ns: Dict[str, StructDecl]


# These are the stack items for the for LR parser stack
Item = Union[
    MemberTypeItem,
    MemberNameItem,
    NamespaceItem,
    StructDecl,
]


class Parser:
    __slots__ = (
        '_tokens',
        '_stack',
    )

    _stack: List[Item]

    def __init__(self, toks: Generator[Lexeme, None, None]):
        self._tokens = toks
        self._stack = list()

    def push(self, item: Item) -> None:
        self._stack.append(item)

    def pop(self) -> Item:
        return self._stack.pop()

    def stacktop(self) -> Item:
        return self._stack[-1]

    def __iter__(self) -> Generator[StructDecl, None, None]:
        self._stack.clear()
        self.push(NamespaceItem({}))

        gen = self._tokens
        st: Optional[State[Parser]] = self._initial

        while st is not None:
            try:
                tok = next(gen)
            except StopIteration:
                raise ParseError('Unexpected end of token-stream')

            st = st(tok)

        ns_item = self.pop()
        assert isinstance(ns_item, NamespaceItem)

        yield from ns_item.ns.values()

    @classmethod
    def fromstring(cls, s: str, file: str = '<string>') -> Parser:
        return cls(lex(s, file=file))

    @classmethod
    def fromfile(cls, p: Path) -> Parser:
        """
        Loads the file, line at a time.
        """

        def gen() -> Generator[Lexeme, None, None]:
            file = str(p)
            with p.open() as f:
                lno = 1
                for line in f:
                    yield from lex(line, file, lno, False)
                    lno += 1
            yield Lexeme('eof', '$', file, lno)
        return cls(gen())

    # Define our states
    @state
    def _initial(self) -> None: ...

    @state
    def _struct_name(self) -> None: ...

    @state
    def _type_start(self) -> None: ...

    @state
    def _member_type(self) -> None: ...

    @state
    def _member_name(self) -> None: ...

    @state
    def _member_end(self) -> None: ...

    # These methods encode what would be the contents of the ACTION/GOTO tables
    # if this were a generated table-based parser.
    #  - The return values define the result of the GOTO table lookup
    #  - The method body either performs a shift or a reduction. Some shifts
    #    are omitted if they are irrelevant to the semantic analysis

    @_initial.on('semicolon')
    def _initial_semi(self, tok: Lexeme) -> Optional[State[Parser]]:
        # reduce
        return self._initial

    @_initial.on('eof')
    def _initial_eof(self, tok: Lexeme) -> Optional[State[Parser]]:
        # accept
        return None

    @_initial.on('kw_struct')
    def _initial_struct_keyword(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        return self._struct_name

    @_type_start.on('lbrace')
    def _type_start_lbrace(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        return self._member_type

    @_struct_name.on('identifier')
    def _struct_name_identifier(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        self.push(StructDecl(tok.val, list()))
        return self._type_start

    @_member_type.on('identifier')
    def _member_type_identifier(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        self.push(MemberTypeItem(tok.val))
        return self._member_name

    @_member_name.on('identifier')
    def _member_name_identifier(self, tok: Lexeme) -> Optional[State[Parser]]:
        # shift
        self.push(MemberNameItem(tok.val))
        return self._member_end

    @_member_end.on('semicolon')
    def _member_end_semicolon(self, tok: Lexeme) -> Optional[State[Parser]]:
        # reduce MemberDecl
        name_item = self.pop()
        type_item = self.pop()

        assert isinstance(name_item, MemberNameItem)
        assert isinstance(type_item, MemberTypeItem)

        member_name = name_item.member_name
        member_type = type_item.member_type

        t = self.stacktop()
        assert isinstance(t, StructDecl)

        t.members.append(MemberDecl(member_name, member_type))
        return self._member_type

    @_member_type.on('rbrace')
    def _member_type_rbrace(self, tok: Lexeme) -> Optional[State[Parser]]:
        # reduce StructDef
        t = self.pop()
        ns = self.stacktop()

        assert isinstance(t, StructDecl)
        assert isinstance(ns, NamespaceItem)

        ns.ns[t.struct_name] = t
        return self._initial


def parse(s: str, file: str = '<string>') -> Generator[StructDecl, None, None]:
    yield from Parser.fromstring(s, file)


def parse_file(p: Path) -> Generator[StructDecl, None, None]:
    yield from Parser.fromfile(p)



"""
xpdt: eXPeditious Data Transfer
"""

from .__version__ import \
    __title__, \
    __description__, \
    __url__, \
    __author__, \
    __author_email__, \
    __copyright__, \
    __license__, \
    __version__  # noqa: F401

from .type import XpdtType
from .typedef import TypeDef
from .basetypes import BaseType, base_types
from .member import MemberDef
from .struct import StructDef
from .decl import MemberDecl, StructDecl
from .namespace import NameSpace
from .parse import parse, parse_file
from .shiftreduce import ParseError
from .load import load

__all__ = (
    'XpdtType',

    'TypeDef',
    'BaseType',
    'base_types',

    'MemberDecl',
    'StructDecl',

    'MemberDef',
    'StructDef',

    'NameSpace',

    'parse',
    'parse_file',

    'ParseError',

    'load',
)



from __future__ import annotations
from typing import Dict, Optional, Callable, Any, TypeVar, Generic, cast
from typing_extensions import Protocol
from functools import wraps

"""
Helpers for shift/reduce parsing
"""

__all__ = (
    'ParseError',
    'state',
)


class Token(Protocol):
    @property
    def tok_type(self) -> Any: ...

    @property
    def val(self) -> str: ...

    @property
    def file(self) -> str: ...

    @property
    def line(self) -> int: ...


class ParseError(Exception):
    tok: Optional[Token]

    def __init__(self,
                 msg: str,
                 /,
                 tok: Optional[Token] = None,
                 ) -> None:
        super().__init__(msg)
        self.tok = tok

    def __str__(self) -> str:
        s = super().__str__()
        tok = self.tok
        if tok is None:
            return s
        return f'{tok.file}:{tok.line}: at "{tok.val}": {s}'


_T = TypeVar('_T')


class State(Generic[_T]):
    __slots__ = ()

    def on(self: _T, *args: str) -> Callable[..., StateFunc[_T]]:
        raise NotImplementedError

    def default(self: _T) -> Callable[..., StateFunc[_T]]:
        raise NotImplementedError

    def __call__(self: _T, tok: Token) -> NextState[_T]:
        raise NotImplementedError


NextState = Optional[State[_T]]
StateFunc = Callable[[_T, Token], NextState[_T]]


def state(func: Callable[[_T], None]) -> State[_T]:
    registry: Dict[Any, StateFunc[_T]] = {}
    dfl: Optional[StateFunc[_T]] = None

    def on(*args: str) -> Callable[..., StateFunc[_T]]:
        def decorator(func: StateFunc[_T]) -> StateFunc[_T]:
            for tok_type in args:
                registry[tok_type] = func
            return func
        return decorator

    def default() -> Callable[..., StateFunc[_T]]:
        nonlocal dfl

        def decorator(func: StateFunc[_T]) -> StateFunc[_T]:
            nonlocal dfl
            dfl = func
            return func

        return decorator

    @wraps(func)
    def wrapper(self: _T, tok: Token) -> NextState[_T]:
        nonlocal dfl

        tok_type = tok.tok_type
        try:
            f = registry[tok_type]
        except KeyError:
            if dfl is not None:
                return dfl(self, tok)

            expected = ', '.join(sorted(registry.keys()))
            raise ParseError(
                f'{tok_type} unexpected, looking for {expected}',
                tok=tok,
            )
        return f(self, tok)

    wrapper.on = on  # type: ignore
    wrapper.default = default  # type: ignore

    return cast(State[_T], wrapper)



import unittest

from xpdt import NameSpace, StructDef, MemberDef, BaseType

class Test_RoundTrips(unittest.TestCase):
    def setUp(self):
        self.code = NameSpace(
            structs=(
                StructDef(
                    name='Item',
                    members=(
                        MemberDef('b', BaseType.bytes),
                        MemberDef('s', BaseType.utf8),
                    ),
                    discriminant=0xfacebeef,
                ),
            ),
            name='test',
        ).gen_dynamic_python()

    def test_item(self):
        orig = self.code.Item(b'\x00' * 4, 'Malkovich')
        b = orig._enum_wrap(ts)
        clone, = self.code.Test.read_many(memoryview(b))
        self.assertEqual((ts, orig), clone)

    def test_item(self):
        bad_utf = b'\xc3\x28'
        ts = 0

        orig = self.code.Item(b'\xff', 'AA')

        a = orig._enum_wrap(ts)
        a = a[:-len(bad_utf)] + bad_utf

        b = orig._enum_wrap(ts)

        first, second = list(self.code.Test.read_many(memoryview(a + b)))
        self.assertIsInstance(first, UnicodeDecodeError)
        self.assertEqual(second, (ts, orig))
